"""Plotting utility functions"""
import pandas as pd
import matplotlib.pyplot as plt

def plot_save_to_file(name, file_name):
    """Save plot to file"""
    plt.title(name)
    plt.savefig(file_name, bbox_inches='tight')

def plot_adm_history_of_systems(systems: pd.DataFrame) -> bool:
    """Generate plot of system ADM levels"""
    if (len(systems)) < 1:
        return False

    systems.created_at = pd.to_datetime(systems.created_at)
    systems.pivot(index='created_at', columns='system_name', values='adm').plot(
        ylim=(1.0, 6.0), xlabel='Date', ylabel='ADM').legend(bbox_to_anchor=(1, 1))

    return True



"""Discord bot collecting and posting ADM summaries."""

import asyncio
import discord
from discord.ext import commands

from pydisadm.bot.adm_cog import Adm
from pydisadm.configuration import Configuration
from pydisadm.controller.adm_controller import AdmController

def create_intents():
    """Create discord intents for bot."""
    intents = discord.Intents.default()
    intents.message_content = True
    return intents

class AdmBot:
    """Discord bot collecting and posting ADM summaries."""

    def __init__(self, configuration: Configuration, controller: AdmController):
        self.configuration = configuration
        self.controller = controller

        intents = create_intents()
        self.bot = commands.Bot(
            application_id=configuration.discord_app_id,
            command_prefix=None,
            intents=intents,
            help_command=None
        )

    async def setup_cogs_async(self):
        """Asynchronously setup cogs"""
        await self.bot.add_cog(Adm(self.bot, self.configuration, self.controller))

    def setup_cogs(self):
        """Setup cogs"""
        return asyncio.run(self.setup_cogs_async())

    async def sync_commands_async(self):
        """Asynchronously synchronize command tree"""
        await self.bot.tree.sync()

    def sync_commands(self):
        """Synchronize command tree"""
        return asyncio.run(self.sync_commands_async())

    def run(self):
        """Run the bot, this function is blocking"""
        self.bot.run(self.configuration.discord_token)



import os

class Configuration:
    """App configuration variables"""

    def __init__(self):
        self.discord_token = os.getenv('DISCORD_TOKEN')
        self.discord_channel = os.getenv('DISCORD_CHANNEL')
        self.discord_app_id = os.getenv('DISCORD_APP_ID')

        alliance_id = os.getenv('ALLIANCE_ID')

        db_keep_adm_days = os.getenv('DB_KEEP_ADM_DAYS', '7')
        if db_keep_adm_days is not None:
            try:
                db_keep_adm_days = int(db_keep_adm_days)
            except ValueError as error:
                raise ValueError(
                    '[configuration] invalid `DB_KEEP_ADM_DAYS` value',
                    db_keep_adm_days) from error

        self.db_keep_adm_days = db_keep_adm_days

        if alliance_id is not None:
            try:
                alliance_id = int(alliance_id)
            except ValueError as error:
                raise ValueError(
                    '[configuration] invalid `ALLIANCE_ID` value',
                    alliance_id) from error

        self.alliance_id = alliance_id

    def __str__(self):
        return (f'{self.__class__.__name__}' +
            f'(discord_token={self.discord_token}, ' +
            f'discord_channel={self.discord_channel}, ' +
            f'discord_app_id={self.discord_app_id}, ' +
            f'alliance_id={self.alliance_id} ' +
            f'db_keep_adm_days={self.db_keep_adm_days})')



"""Controller implementation for ADM operations"""
import os
import numpy as np
import pandas as pd
from tabulate import tabulate

from pydisadm.configuration import Configuration
from pydisadm.services.database import Database
from pydisadm.services.esi import sovereignty_structures
from pydisadm.utils.adm_utils import adm_from_index
from pydisadm.utils.plot_utils import plot_adm_history_of_systems, plot_save_to_file


class AdmController:
    """Controller implementation for ADM operations"""

    def __init__(self, configuration: Configuration, database: Database):
        self.configuration = configuration
        self.database = database

    def generate_tier_list(self):
        """Generate a system tier list based on tier values"""
        system_adms = self.database.select_systems()

        system_adms.sort_values(by='adm', inplace=True, ascending=True)

        s_tier = system_adms.loc[system_adms['tier'] == 's_tier']
        a_tier = system_adms.loc[system_adms['tier'] == 'a_tier']
        b_tier = system_adms.loc[system_adms['tier'] == 'b_tier']
        c_tier = system_adms.loc[system_adms['tier'] == 'c_tier']
        d_tier = system_adms.loc[system_adms['tier'] == 'd_tier']

        sorted_summary = [
            {'Tier': 'D', 'Systems': '\n'.join(d_tier['solarSystemName']),
             'ADM': '\n'.join(d_tier['adm'].map(str)),
             'Constellation': '\n'.join(d_tier['constellationName']),
             'Region': '\n'.join(d_tier['regionName'])},
            {'Tier': 'C', 'Systems': '\n'.join(c_tier['solarSystemName']),
             'ADM': '\n'.join(c_tier['adm'].map(str)),
             'Constellation': '\n'.join(c_tier['constellationName']),
             'Region': '\n'.join(c_tier['regionName'])},
            {'Tier': 'B', 'Systems': '\n'.join(b_tier['solarSystemName']),
             'ADM': '\n'.join(b_tier['adm'].map(str)),
             'Constellation': '\n'.join(b_tier['constellationName']),
             'Region': '\n'.join(b_tier['regionName'])},
            {'Tier': 'A', 'Systems': '\n'.join(a_tier['solarSystemName']),
             'ADM': '\n'.join(a_tier['adm'].map(str)),
             'Constellation': '\n'.join(a_tier['constellationName']),
             'Region': '\n'.join(a_tier['regionName'])},
            {'Tier': 'S', 'Systems': '\n'.join(s_tier['solarSystemName']),
             'ADM': '\n'.join(s_tier['adm'].map(str)),
             'Constellation': '\n'.join(s_tier['constellationName']),
             'Region': '\n'.join(s_tier['regionName'])}
        ]

        table = tabulate(sorted_summary, showindex=False, headers='keys',
                         tablefmt='fancy_grid', numalign='left', stralign='center')

        generated_at = self.database.select_most_recent_row()

        return (table, generated_at['created_at'][0])

    def create_history_graph(self, name, file_name):
        """Generate an ADM history graph and save to file"""
        systems_by_name = self.database.select_system_by_name(name)

        if plot_adm_history_of_systems(systems_by_name):
            plot_save_to_file('ADM History', file_name)

        return os.path.isfile(file_name)

    def create_spreadsheet(self, file_name):
        """Generate a spreadsheet of ADM data and save to file"""
        system_adms = self.database.select_systems()
        system_adms.sort_values(by='adm', inplace=True, ascending=False)

        csv_columns = [
            'system_id',
            'adm',
            'tier',
            'created_at',
            'solarSystemName',
            'constellationName',
            'regionName'
        ]

        system_adms.to_csv(file_name, index=False, columns=csv_columns)

        return os.path.isfile(file_name)

    def tier_list(self, systems):
        """Assign tier column based on system ADM"""
        conditions = [
            (systems['adm'] >= 6.0),
            (systems['adm'] >= 5.0) & (systems['adm'] < 6.0),
            (systems['adm'] >= 4.0) & (systems['adm'] < 5.0),
            (systems['adm'] >= 3.5) & (systems['adm'] < 4.0),
            (systems['adm'] < 3.5)
        ]

        values = [
            's_tier',
            'a_tier',
            'b_tier',
            'c_tier',
            'd_tier'
        ]

        systems['tier'] = np.select(conditions, values)

        return systems

    def create_system_adm_from_index(self, system, military: int, industrial: int, strategic: int):
        """Create a DataFrame for system ADM data from index values"""
        adm = adm_from_index(military, industrial, strategic)

        return self.create_system_adm(system, adm)


    def create_system_adm(self, system, adm):
        """Create a DataFrame for system ADM data"""
        system_adm = pd.DataFrame(columns=['system_id', 'adm'])
        system_adm['system_id'] = system['solarSystemID']
        system_adm['adm'] = [adm]

        self.tier_list(system_adm)

        return system_adm

    def get_alliance_structures(self, structures, alliance_id):
        """Retrieve a list of alliance sovreignty structures"""
        structure_data = pd.DataFrame.from_dict(structures)

        drop_columns = [
            'structure_id',
            'structure_type_id',
            'vulnerable_end_time',
            'vulnerable_start_time',
        ]

        structure_data.drop(columns=drop_columns, inplace=True)

        return structure_data[structure_data['alliance_id'] == alliance_id]

    def update_system_adm(self, system_name: str, adm: float) -> bool:
        """Update ADM for system with name"""
        system = self.database.select_system_with_name(system_name)

        if system.empty:
            return False

        insert_systems = self.create_system_adm(system, adm)

        self.database.insert_systems(insert_systems)

        return True

    def update_system_adm_from_index(self,
                                     system_name: str,
                                     military: int,
                                     industrial: int,
                                     strategic: int) -> bool:
        """Update ADM for system with name from index values"""
        system = self.database.select_system_with_name(system_name)

        if system.empty:
            return (False, 0)

        insert_systems = self.create_system_adm_from_index(
            system, military, industrial, strategic)

        self.database.insert_systems(insert_systems)

        return (True, insert_systems['adm'][0])

    def get_system_adms(self, alliance_id):
        """Retrieve ADM for systems controlled by alliance"""
        structures = sovereignty_structures()
        system_adms = self.get_alliance_structures(structures, alliance_id)

        system_adms.drop_duplicates(inplace=True)

        drop_columns = ['alliance_id']
        system_adms.drop(columns=drop_columns, inplace=True)

        rename_columns = {'solar_system_id': 'system_id',
                          'vulnerability_occupancy_level': 'adm'}
        system_adms.rename(columns=rename_columns, inplace=True)

        self.tier_list(system_adms)

        return system_adms

    def get_recommended_system(self):
        """Retrieve recommended system to raise ADM"""
        systems = self.database.select_systems()
        systems.sort_values(by='adm', inplace=True, ascending=True)

        return systems.iloc[0]

    def update_adm_data(self):
        """Update ADM data"""
        alliance_id = self.configuration.alliance_id
        system_adms = self.get_system_adms(alliance_id)

        self.database.insert_systems(system_adms)

    def purge_adm_records(self, days_old: int):
        """Purge adm records more than days_old days old"""

        self.database.delete_system_rows(days_old)

    def write_file(self, file_name, content) -> bool:
        """Write content to file"""
        with open(file_name, 'w', encoding='UTF-8') as file:
            file.write(content)

        return os.path.isfile(file_name)

    def delete_file(self, file_name):
        """Delete file"""
        os.remove(file_name)



"""Database service"""
import sqlite3
import pandas as pd

class Database:
    """Database service implementation"""

    def __init__(self):
        self.conn = sqlite3.connect('adm-data.sqlite')
        self.setup()

    def setup(self):
        """Setup database schema"""
        cur = self.conn.cursor()
        cur.execute("""
                        CREATE TABLE IF NOT EXISTS systems (
                            id INTEGER PRIMARY KEY AUTOINCREMENT, 
                            system_id INTEGER NOT NULL, 
                            adm REAL NOT NULL, 
                            tier TEXT NOT NULL,
                            created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
                        )
                    """)

        cur.execute("""
                        CREATE TABLE IF NOT EXISTS map (
                            solarSystemID INTEGER PRIMARY KEY NOT NULL, 
                            constellationID INTEGER NOT NULL, 
                            regionID INTEGER NOT NULL, 
                            solarSystemName TEXT NOT NULL,
                            constellationName TEXT NOT NULL,
                            regionName TEXT NOT NULL
                        )
                    """)
        self.conn.commit()

    def insert_systems(self, systems):
        """Insert system ADM records"""
        systems.to_sql('systems', self.conn, index=False, if_exists='append')

        self.conn.commit()

    def insert_map_data(self, map_data):
        """Insert map data"""
        map_data.to_sql('map', self.conn, index=True, if_exists='replace')

        self.conn.commit()

    def select_system_with_name(self, system_name) -> pd.DataFrame:
        """Select system matching name"""
        return pd.read_sql_query("SELECT * FROM map WHERE solarSystemName = ?",
                                 self.conn, params=[system_name])

    def select_most_recent_row(self) -> pd.DataFrame:
        """Select the most recent ADM record"""
        return pd.read_sql_query("SELECT created_at FROM systems ORDER BY created_at DESC LIMIT 1",
                                 self.conn)

    def select_systems(self) -> pd.DataFrame:
        """Select most recent record of all systems"""
        return pd.read_sql_query("""
            SELECT t1.*, t2.solarSystemName, t2.constellationName, t2.regionName FROM systems t1 
            LEFT JOIN map t2 ON t1.system_id = t2.solarSystemID 
            WHERE t1.created_at = (SELECT MAX(t3.created_at) FROM systems t3 WHERE t3.system_id = t1.system_id);
        """, self.conn)

    def select_system_by_name(self, name) -> pd.DataFrame:
        """Select systems by name"""
        sql = """
            SELECT map.solarSystemName system_name, adm, tier, created_at FROM systems
            INNER JOIN map ON map.solarSystemID = systems.system_id
            WHERE map.solarSystemName=? OR map.constellationName=? OR map.regionName=? ORDER BY created_at
        """
        return pd.read_sql_query(sql, self.conn, params=[name, name, name])

    def select_system_history(self, system, limit) -> pd.DataFrame:
        """Select history of a single system"""
        return pd.read_sql_query("""
            SELECT system_id, adm, tier, created_at FROM systems
            INNER JOIN map ON map.solarSystemID = systems.system_id
            WHERE map.solarSystemName=? ORDER BY created_at DESC LIMIT ?
        """, self.conn, params=[system, limit])

    def delete_system_rows(self, days_old):
        """Delete system rows older than days_old"""
        cur = self.conn.cursor()
        cur.execute(f"""
            DELETE FROM systems 
            WHERE id IN (
                SELECT id FROM systems WHERE created_at < (select datetime('now', '-{days_old} day'))
            )
        """)

        self.conn.commit()



#!/usr/bin/env python

import signal
import sys
import threading

from dotenv import load_dotenv

from pydisadm.bot.adm_bot import AdmBot
from pydisadm.configuration import Configuration
from pydisadm.controller.adm_controller import AdmController
from pydisadm.loader.static_data import update_static_data
from pydisadm.runnable.runnable_refresh import run_auto_refresh
from pydisadm.services.database import Database

interrupt_event = threading.Event()

def signal_handler(_1, _2):
    print('Interrupted by CTRL+C')
    interrupt_event.set()
    sys.exit(0)


def main() -> int:
    load_dotenv(verbose=True)

    signal.signal(signal.SIGINT, signal_handler)

    run_auto_refresh(interrupt_event)

    configuration = Configuration()

    database = Database()

    update_static_data(database)

    controller = AdmController(configuration, database)
    
    controller.update_adm_data()
    controller.purge_adm_records(configuration.db_keep_adm_days)

    bot = AdmBot(configuration, controller)
    bot.setup_cogs()
    bot.run()

    interrupt_event.set()

    return 0

if __name__ == '__main__':
    sys.exit(main())



"""Module for running scheduled refresh"""
import logging
import schedule

from pydisadm.configuration import Configuration
from pydisadm.controller.adm_controller import AdmController
from pydisadm.services.database import Database
from pydisadm.utils.thread_utils import run_threaded

logger = logging.getLogger('adm_auto_refresh')

def scheduler_loop(interrupt_event):
    """Scheduler main loop"""
    configuration = Configuration()
    database = Database()

    controller = AdmController(configuration, database)

    schedule.every().day.at('11:30', tz='UTC').do(refresh_job, controller, configuration)

    while True:
        schedule.run_pending()
        if interrupt_event.wait(timeout=1000):
            break

def refresh_job(controller: AdmController, configuration: Configuration):
    """Refresh adm data"""
    logger.info('updating adm data...')
    controller.update_adm_data()
    logger.info('adm data update finished')

    logger.info('purge old adm data...')
    controller.purge_adm_records(configuration.db_keep_adm_days)
    logger.info('purge finished')

def run_auto_refresh(interrupt_event):
    """Run automatic adm data refresh in separate thread"""
    logger.info('starting adm data update thread')
    run_threaded(lambda: scheduler_loop(interrupt_event))



"""Utilities for loading static data"""
import pandas as pd

from pydisadm.loader.datasets import get_constellations, get_regions, get_solar_systems

def load_solar_systems() -> pd.DataFrame:
    """Load solar systems data"""
    solar_systems = pd.read_csv(get_solar_systems(), usecols=[
        'regionID','constellationID','solarSystemID','solarSystemName'])
    solar_systems.set_index('solarSystemID', inplace=True)

    constellations = pd.read_csv(get_constellations(), usecols=[
        'constellationID','constellationName'])
    constellations.set_index('constellationID', inplace=True)

    regions = pd.read_csv(get_regions(), usecols=['regionID','regionName'])
    regions.set_index('regionID', inplace=True)

    solar_systems = solar_systems.join(constellations, on='constellationID', how='left')
    solar_systems = solar_systems.join(regions, on='regionID', how='left')

    return solar_systems

def update_static_data(database):
    """Populate database with solar systems data"""
    solar_systems = load_solar_systems()
    database.insert_map_data(solar_systems)
